/*
 * Copyright (c) 2012-2014 Wind River Systems, Inc.
 *
 * SPDX-License-Identifier: Apache-2.0
 */

#include <zephyr/kernel.h>
#include <helpers/nrfx_gppi.h>
#include <nrfx_timer.h>
#include <nrfx_gpiote.h>
#include <assert.h>

#define TIMER_INST_IDX 1

static const nrfx_timer_t timer = NRFX_TIMER_INSTANCE(TIMER_INST_IDX);
const uint32_t rising_edge_pin = 28;
const uint32_t falling_edge_pin = 29;

static void gpiote_event_handler(nrfx_gpiote_pin_t pin,
				 nrfx_gpiote_trigger_t action, void *unused)
{
	if (pin == falling_edge_pin) {
		/* Read out timer CC register and clear it */
		uint32_t timer_cc_val = nrf_timer_cc_get(timer.p_reg, NRF_TIMER_CC_CHANNEL0);
		printk("Raw timer val: %x\n", timer_cc_val);
		
		/* Calculate pulse length, 16M base freq */
		uint32_t pulse_len_us = (timer_cc_val >> 4) - 1;
		uint32_t pulse_len_ms = pulse_len_us / 1000;
		printk("ms: %d\n", pulse_len_ms);
		printk("us: %d\n", pulse_len_us);

		/* Clear the timer, as all calculations are based on starting from 0 */
		nrfx_timer_clear(&timer);
	} else {
		printk("Unknown pin.\n");
	}
}

static void gpiote_init(void)
{
	uint8_t rising_edge_in_ch, falling_edge_in_ch;
	uint32_t err = nrfx_gpiote_init(7);
	__ASSERT_NO_MSG(err == NRFX_SUCCESS);

	err = nrfx_gpiote_channel_alloc(&rising_edge_in_ch);
	if (err != NRFX_SUCCESS) {
		printk("Failed to allocate in_channel, error: 0x%08X", err);
	}
	
	err = nrfx_gpiote_channel_alloc(&falling_edge_in_ch);
	if (err != NRFX_SUCCESS) {
		printk("Failed to allocate in_channel, error: 0x%08X", err);
	}	

	/* Initialize input pin to generate event on high to low transition
	 * (falling edge) and call button_handler()
	 */
	static const nrfx_gpiote_input_config_t rising_input_config = {
		.pull = NRF_GPIO_PIN_PULLDOWN,
	};
	const nrfx_gpiote_trigger_config_t rising_trigger_config = {
		.trigger = NRFX_GPIOTE_TRIGGER_LOTOHI,
		.p_in_channel = &rising_edge_in_ch,
	};
	static const nrfx_gpiote_input_config_t falling_input_config = {
		.pull = NRF_GPIO_PIN_PULLDOWN,
	};
	const nrfx_gpiote_trigger_config_t falling_trigger_config = {
		.trigger = NRFX_GPIOTE_TRIGGER_HITOLO,
		.p_in_channel = &falling_edge_in_ch,
	};	
	static const nrfx_gpiote_handler_config_t handler_config = {
		.handler = gpiote_event_handler,
	};

	/* Do not generate ISR for rising, we only want for falling */
	err = nrfx_gpiote_input_configure(rising_edge_pin,
					  &rising_input_config,
					  &rising_trigger_config,
					  NULL);
	__ASSERT_NO_MSG(err == NRFX_SUCCESS);					  
	
	err = nrfx_gpiote_input_configure(falling_edge_pin,
					  &falling_input_config,
					  &falling_trigger_config,
					  &handler_config);
	__ASSERT_NO_MSG(err == NRFX_SUCCESS);

	nrfx_gpiote_in_event_enable(falling_edge_pin, 1);
	nrfx_gpiote_in_event_enable(rising_edge_pin, 1);	
}

static void gppi_init(void)
{
	uint8_t gppi_channel1, gppi_channel2;
    uint32_t status = nrfx_gppi_channel_alloc(&gppi_channel1);
    __ASSERT_NO_MSG(status == NRFX_SUCCESS);
    
	status = nrfx_gppi_channel_alloc(&gppi_channel2);
    __ASSERT_NO_MSG(status == NRFX_SUCCESS);	

    nrfx_gppi_channel_endpoints_setup(gppi_channel1,
        nrfx_gpiote_in_event_addr_get(rising_edge_pin),
        nrfx_timer_task_address_get(&timer, NRF_TIMER_TASK_START));
    
	nrfx_gppi_channel_endpoints_setup(gppi_channel2,
        nrfx_gpiote_in_event_addr_get(falling_edge_pin),
        nrfx_timer_task_address_get(&timer, NRF_TIMER_TASK_CAPTURE0));

	nrfx_gppi_fork_endpoint_setup(gppi_channel2,
		nrf_timer_task_address_get(timer.p_reg, NRF_TIMER_TASK_STOP));

    nrfx_gppi_channels_enable(BIT(gppi_channel1));
	nrfx_gppi_channels_enable(BIT(gppi_channel2));

}

static void timer_init(void)
{
    nrfx_timer_config_t timer_config = NRFX_TIMER_DEFAULT_CONFIG;
    timer_config.bit_width = NRF_TIMER_BIT_WIDTH_32;

    uint32_t status = nrfx_timer_init(&timer, &timer_config, NULL);
    __ASSERT_NO_MSG(status == NRFX_SUCCESS);
}

int main(void)
{
	printk("Lets time signals! On board %s\n", CONFIG_BOARD);
#if defined(__ZEPHYR__)
    IRQ_DIRECT_CONNECT(NRFX_IRQ_NUMBER_GET(NRF_TIMER_INST_GET(TIMER_INST_IDX)), IRQ_PRIO_LOWEST,
                       NRFX_TIMER_INST_HANDLER_GET(TIMER_INST_IDX), 0);
    IRQ_DIRECT_CONNECT(NRFX_IRQ_NUMBER_GET(NRF_GPIOTE), IRQ_PRIO_LOWEST, nrfx_gpiote_irq_handler,
                       0);
#endif

	timer_init();
	gpiote_init();
	gppi_init();

	return 0;
}
